package inria.meije.rc.sugarcubes;

import java.util.*;

public class EventMachine extends Cube implements Domain, Machine
{
    protected boolean move = false, endOfInstant = false, beginingOfInstant = true;
    protected int instant = 1;
    protected Hashtable frozenStore = new Hashtable(), addToShell = new Hashtable()
                      , shellEnv = new Hashtable();
    protected Vector toFreeze = new Vector(), presentEvents = new Vector();
    protected Link currentLink = this;
    protected Instruction addToProgram = null;

    public EventMachine(){ super("noname", new Nothing()); }
    public EventMachine(Instruction i){ super("noname",i); }
    public EventMachine(JavaStringExpression n,JavaObjectExpression obj,Instruction i
               ,JavaInstruction fin,JavaInstruction f,JavaInstruction w){
        super(n,obj,i,fin,f,w);
    }

    public void setCurrentLink(Link link){ currentLink = link; }
    public Link currentLink(){ return currentLink; }
    public void stopReached(){}
    public void add(Instruction inst){
        synchronized(this){
            if(addToProgram!=null)
                addToProgram = new Merge(addToProgram,inst);
            else
                addToProgram = inst;
        }
    }
    public void addToShell(String name,Instruction inst){
        Instruction old = (Instruction)addToShell.get(name);
        if (old==null)
            addToShell.put(name,inst);
        else
            addToShell.put(name,new Merge(old,inst));
    }
    protected void processAddToShell(){
        if (addToShell.isEmpty()) return;
        Enumeration list = addToShell.keys();
        while (list.hasMoreElements()){
            String name = (String)list.nextElement();
            Shell shell = (Shell)shellEnv.get(name);
            if (shell==null)
                System.err.println("unknown shell: "+name);
            else
                shell.add((Instruction)addToShell.get(name));
        }
        addToShell.clear();
    }
    public void registerShell(String name,Shell shell){ shellEnv.put(name,shell); }
    public void removeShell(String name){ shellEnv.remove(name); }
    public int currentInstant(){ return instant; }
    public void newMove(){ move = true; }
    protected void newInstant(){ instant++; presentEvents.removeAllElements(); }
    protected void notifyBeginOfInstantToJava(){}
    protected byte activation(Context context){
        move = false;
        if(beginingOfInstant){
            synchronized(this){
                if(addToProgram != null){ super.add(addToProgram); addToProgram = null; }
            }
            notifyBeginOfInstantToJava();
            processAddToShell();
            beginingOfInstant = false;
        }
        byte res = super.activation(this);
        if (res == STOP){
            endOfInstant = false;
            newInstant();
            beginingOfInstant = true;
            processFreezeOrders();
            notifyEndOfInstantToJava();
        }
        else if(res == SUSP){
            if (!move) endOfInstant = true;
            context.newMove();
        }
        return res;
    }
    protected void notifyEndOfInstantToJava(){}
    public boolean isEndOfInstant(){ return endOfInstant; }
    protected Instruction rest(){
        return new EventMachine(new JavaStringValue(name()),new JavaObjectValue(javaObject()),body().residual(),onTerminate,onFreeze,onWarmUp);
    }
    public String toString(){
        return "machine "+name()+" "+program()+" end";
    }
    public void freeze(Context context){ doFreeze(context); }
    public void freezeOrder(String name){ toFreeze.addElement(name); }
    public boolean isToBeFrozen(String name){ return toFreeze.contains(name); }
    protected void processFreezeOrders(){
        if (toFreeze.isEmpty()) return;
        body().freeze(this);
        toFreeze.removeAllElements();
    }
    public void storeFrozenInstruction(String name,Instruction frozen){
        frozenStore.put(name,frozen);
    }
    public Instruction getFrozenInstruction(String name){
        return (Instruction) frozenStore.remove(name);
    }
    public boolean react(){
        byte res = SUSP;
        while (res == SUSP) res = activ(this);
        return res == TERM ? true : false;
    }
    public Instruction program(){
    	return body();
    }
    public boolean isGenerated(String name){
        return presentEvents.contains(name);
    }
    public void generate(String name){
        if(!presentEvents.contains(name)) presentEvents.addElement(name);
    }
    public boolean swapEventPresence(String name,boolean presence){
        boolean present = presentEvents.contains(name);
    	if(presence){
            if(!present) presentEvents.addElement(name);
    	}
        else
            if(present) presentEvents.removeElement(name);
        return present;
    }
    public Object clone(){
        EventMachine inst = (EventMachine)super.clone();
        inst.frozenStore = new Hashtable();
        inst.addToShell = new Hashtable();
        inst.shellEnv = new Hashtable();
        inst.toFreeze = new Vector();
        inst.presentEvents = new Vector();
        inst.currentLink = inst;
        inst.addToProgram = (Instruction) addToProgram.clone();
        return inst;
    }
}
