package inria.meije.rc.sugarcubes;

public class Until extends BinaryInstruction
{
    protected Config config;
    protected boolean activeHandle = false;
    protected boolean resumeBody = true;

    public Until(Config config,Instruction body,Instruction handler){ 
        this.config = config; left = body; right = handler;
    }
    public Until(Config config,Instruction body){ 
        this(config,body,new Nothing());
    }
    public Until(String name,Instruction body,Instruction handler){ 
        this(new PosConfig(name),body,handler);
    }
    public Until(String name,Instruction body){ 
        this(name,body,new Nothing());
    }

    public String toString(){ 
        if (right instanceof Nothing) return "do "+left+" until "+config;
        return "do "+left+" until "+config+" actual "+right+" end"; 
    }
    public Object clone(){
        Until inst = (Until)super.clone();
        inst.config = (Config)config.clone();
        return inst;
    }
    protected Instruction rest(){ 
        if (activeHandle) return right.residual();
        return new Until(config,left.residual(),right);
    }
    protected byte activation(Context context){ 
        Domain domain = (Domain)context;
        if (activeHandle)
            return right.activ(domain);
        if (resumeBody){
            byte res = left.activ(domain);
            if (res != STOP) return res;
            resumeBody = false; 
        }
        if (!config.fixed(domain)) return SUSP; 
        if (config.evaluate(domain)){
            activeHandle = true;
            left.notifyTerminationToJava(context);
            if (domain.isEndOfInstant()) return STOP;
            return right.activ(domain);
        }
        resumeBody = true;
        return STOP;
    }
    public void freeze(Context context){
        if(activeHandle)
            right.freeze(context);
        else
            left.freeze(context);
    }
    public void notifyTerminationToJava(Context context){
        if(activeHandle)
            right.notifyTerminationToJava(context);
        else
            left.notifyTerminationToJava(context);
    }
    public void notifyWarmUpToJava(Context context){
        if(activeHandle)
            right.notifyWarmUpToJava(context);
        else
            left.notifyWarmUpToJava(context);
    }
    public void notifyFreezeToJava(Context context){
        if(activeHandle)
            right.notifyFreezeToJava(context);
        else
            left.notifyFreezeToJava(context);
    }
}
